#!/usr/bin/python
import pysam
import re
import sys
import time
import os
import multiprocessing as mp
import queue

from classes import read as r
from classes import segment as s
from classes import variant as v

sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir))
import NanoSV

segments = {}
variants = {}
reads = {}
coverages = []

def parse_bam():
    """
    Reads bam file and saves reads and their segments in objects of the Read en Segment classes.
    :param bamfile used to open bam file:
    """
    global sample_name, header, bam

    sys.stderr.write(time.strftime("%c") + " Busy with parsing bam file...\n")
    bam = pysam.AlignmentFile(NanoSV.opts_bam, 'rb')
    header = bam.header
    if not bam.has_index():
        sys.exit('The bam has no index file')
    if 'HD' in header:
        if not header['HD']['SO'] == 'coordinate':
            sys.exit('The bam file is not coordinate sorted')
    if 'RG' in header:
        if type(header['RG']) is list:
            sample_name = header['RG'][0]['SM']
        else:
            sample_name = header['RG']['SM']
    else:
        sample_name = re.sub('(\.sorted)?\.bam$', '', str(NanoSV.opts_bam))

    contig_list = []
    for contig_dict in header['SQ']:
        contig_list.append(contig_dict['SN'])

    q = mp.Queue()
    q_out = mp.Queue()
    for contig in contig_list:
        q.put(contig)

    processes = [mp.Process(target=parse_chr_bam, args=(q, q_out, NanoSV.opts_bam)) for x in range(NanoSV.opts_threads)]

    for p in processes:
        p.start()

    liveprocs = list(processes)
    while liveprocs:
        try:
            while 1:
                contig_segments, contig_variants = q_out.get(block=False, timeout=1)
                segments.update(contig_segments)
                variants.update(contig_variants)
        except queue.Empty:
            pass

        time.sleep(0.5)    # Give tasks a chance to put more data in
        if not q.empty():
            continue
        liveprocs = [p for p in liveprocs if p.is_alive()]

    for p in processes:
        p.join()

    for contig in segments:
        for pos in segments[contig]:
            for id in segments[contig][pos]:
                segment = segments[contig][pos][id]
                if segment.qname in reads:
                    read = read = reads[segment.qname]
                else:
                    read = r.Read(segment.qname, segment.rlength)
                    reads[segment.qname] = read
                read.addSegment(segment)
    write_bed()
    bam.close()


def parse_chr_bam(q, q_out, bamfile):
    """
    Reachs each alignment for each chromosome. Separate for multiprocessing
    """
    global segments_to_check, segments, variants, tmp_variants, F

    while True:

        try:
            contig = q.get(block=False, timeout=1)
        except queue.Empty:
            break

        tmp_variants = {}
        segments = {contig: {}}

        segments_to_check = {}
        variants = {}

        for item in header["SQ"]:
            if item['SN'] == contig:
                variants[item['SN']] = {}
                for bin in range(0,int(item['LN']/NanoSV.opts_variant_bin_size)):
                    variants[item['SN']][bin] = {}

        F=pysam.AlignmentFile(bamfile, 'rb')
        Alignment = F.fetch(contig, multiple_iterators=True)

        previous_refname = -1
        previous_cursor = -1
        for line in Alignment:
            if line.flag & 4:
                continue
            remove = [qname_clip for qname_clip in segments_to_check if line.reference_start > segments_to_check[qname_clip][1]]
            for q_clip in remove: del segments_to_check[q_clip]
            clip, clip_2 = calculate_clip(line)
            created_subsegments = search_for_indels(line, clip, clip_2)
            for segment in created_subsegments:
                if not int(segment.pos) in segments[segment.rname]:
                    segments[segment.rname][segment.pos] = {}

                segments[segment.rname][segment.pos][segment.id[-1]] = segment
                segments_to_check[segment.id[-1]] = [segment.pos, segment.end]
                previous_refname = line.reference_name
            if NanoSV.opts_phasing_on and line.seq and line.query_qualities:
                if line.has_tag('MD'):
                    if not NanoSV.opts_snp_file:
                        find_variations_md(line.cigartuples, line.reference_name, line.reference_start + 1, line.query_qualities, line.seq, line.get_tag('MD'), created_subsegments)
                else:
                    if not NanoSV.opts_snp_file:
                        find_variations_cigar(line.cigartuples, line.reference_name, line.reference_start + 1, line.query_qualities, line.seq, created_subsegments)
            if previous_cursor == -1:
                previous_cursor = line.reference_start
            if NanoSV.opts_phasing_on and line.seq and line.query_qualities and line.reference_start > previous_cursor:
                remove_variations(previous_cursor, line.reference_start, line.reference_name)
            previous_cursor = line.reference_start
        if NanoSV.opts_phasing_on and NanoSV.opts_snp_file:
            read_snp_vcf(contig)
        F.close()

        q_out.put( (segments, variants) )

def calculate_pid(line, query_alignment_length):
    """
    calculates percentage identity of alignment of read
    :param line:
    :param query_alignment_length:
    :return pid:
    """
    if line.has_tag('MD'):
        matches = sum(map(int, re.findall(r"(\d+)", line.get_tag('MD'))))
        pid = format(matches / query_alignment_length, '.3f')
    else:
        pid = format(line.get_cigar_stats()[0][7] / query_alignment_length, '.3f')
        if pid == "0.000":
            pid = format(line.get_cigar_stats()[0][0] / query_alignment_length, '.3f')
    return pid

def calculate_clip(line):
    """
    calculates clip of segment
    :param line:
    :return list with clip of both sides of the segment:
    """
    if line.flag & 16:
        if line.cigartuples[-1][0] == 5 or line.cigartuples[-1][0] == 4:
            clip = line.cigartuples[-1][1]
        else:
            clip = 0
        if line.cigartuples[0][0] == 5 or line.cigartuples[0][0] == 4:
            clip_2 = line.cigartuples[0][1]
        else:
            clip_2 = 0
    else:
        if line.cigartuples[0][0] == 5 or line.cigartuples[0][0] == 4:
            clip = line.cigartuples[0][1]
        else:
            clip = 0
        if line.cigartuples[-1][0] == 5 or line.cigartuples[-1][0] == 4:
            clip_2 = line.cigartuples[-1][1]
        else:
            clip_2 = 0
    return [clip, clip_2]

def create_pattern():
    """
    creates a pattern to search for deletions and indels in cigar string using regular expressions
    :return pattern to search for with re.finditer():
    """
    pattern_list = [''] * len(NanoSV.opts_min_indel_size)
    for i in range(0, len(NanoSV.opts_min_indel_size)):
        j = len(NanoSV.opts_min_indel_size) - i - 1
        if j > 0:
            pattern_list[i] += '[' + str(int(NanoSV.opts_min_indel_size[i]) + 1) + '-9]\d{' + str(j) + ',}'
        for k in range(i + 1, len(pattern_list)):
            pattern_list[k] += '[' + str(int(NanoSV.opts_min_indel_size[i])) + '-9]'
    pattern_list[-1] += '[' + str(int(NanoSV.opts_min_indel_size[-1])) + '-9]'
    pattern_list.append("\d{" + str(len(NanoSV.opts_min_indel_size) + 1) + ",}")
    pattern = re.compile(r'' + "(" + "|".join(pattern_list) + ")([DI])" + '')
    return pattern

def search_for_indels(line, clip, clip_2):
    """
    Looks for indels that could be broken up to create breakpoints. Breakpoints are detected by NanoSV
    (Gapped alignment fix)
    :param line:
    :param clip:
    :param clip_2:
    :return list of created subsegments:
    """
    cigarstring = line.cigarstring
    reference_start = line.reference_start + 1
    line_query_alignment_length = line.query_alignment_length
    query_alignment_length = 0
    c_start = 0
    pattern = create_pattern()
    i = 0
    created_subsegments = []
    for m in re.finditer(pattern, line.cigarstring):
        line.cigarstring = cigarstring[c_start:m.start(0)]
        reference_end = reference_start + line.get_cigar_stats()[0][0] + line.get_cigar_stats()[0][2] + \
                        line.get_cigar_stats()[0][7] + line.get_cigar_stats()[0][8] - 1
        if line.flag & 16:
            clip_2 = clip_2 + query_alignment_length
            query_alignment_length = line.get_cigar_stats()[0][0] + line.get_cigar_stats()[0][1] + \
                                     line.get_cigar_stats()[0][7] + line.get_cigar_stats()[0][8]
            if i == 0:
                clip = clip + line_query_alignment_length - query_alignment_length
            else:
                clip = clip - query_alignment_length
            i += 1
        else:
            clip = clip + query_alignment_length

            query_alignment_length = line.get_cigar_stats()[0][0] + line.get_cigar_stats()[0][1] + \
                                     line.get_cigar_stats()[0][7] + line.get_cigar_stats()[0][8]
            if i == 0:
                clip_2 = clip_2 + line_query_alignment_length - query_alignment_length
            else:
                clip_2 = clip_2 - query_alignment_length
            i += 1
        if query_alignment_length == 0:
            reference_start = int(reference_end) + 1
            if line.flag & 16:
                if m.group(2) == "I":
                    clip -= int(m.group(1))
            else:
                if m.group(2) == "I":
                    clip += int(m.group(1))
            if m.group(2) == "D":
                reference_start += int(m.group(1))
            c_start = m.end(0)
            continue
        pid = keep_segment(line, query_alignment_length)
        if not pid:
            reference_start = int(reference_end) + 1
            if line.flag & 16:
                if m.group(2) == "I":
                    clip -= int(m.group(1))
            else:
                if m.group(2) == "I":
                    clip += int(m.group(1))
            if m.group(2) == "D":
                reference_start += int(m.group(1))
            c_start = m.end(0)
            continue
        segment = s.Segment(line.query_name, line.flag, line.reference_name, reference_start, line.mapping_quality, query_alignment_length, clip, clip_2, pid, line.infer_read_length())
        segment.end = reference_end
        created_subsegments.append(segment)
        reference_start = int(reference_end) + 1
        if line.flag & 16:
            if m.group(2) == "I":
                clip -= int(m.group(1))
        else:
            if m.group(2) == "I":
                clip += int(m.group(1))
        if m.group(2) == "D":
            reference_start += int(m.group(1))
        c_start = m.end(0)
    line.cigarstring = cigarstring[c_start:]
    reference_end = reference_start + line.get_cigar_stats()[0][0] + line.get_cigar_stats()[0][2] + \
                    line.get_cigar_stats()[0][7] + line.get_cigar_stats()[0][8] - 1
    if line.flag & 16:
        if i != 0:
            clip_2 = clip_2 + query_alignment_length
            query_alignment_length = line.get_cigar_stats()[0][0] + line.get_cigar_stats()[0][1] + line.get_cigar_stats()[0][7] + line.get_cigar_stats()[0][8]
            clip = clip - query_alignment_length
        else:
            query_alignment_length = line_query_alignment_length
    else:
        if i != 0:
            clip = clip + query_alignment_length
            query_alignment_length = line.get_cigar_stats()[0][0] + line.get_cigar_stats()[0][1] + \
                                     line.get_cigar_stats()[0][7] + line.get_cigar_stats()[0][8]
            clip_2 = clip_2 - query_alignment_length
        else:
            query_alignment_length = line_query_alignment_length
    pid = keep_segment(line, query_alignment_length)
    if pid:
        segment = s.Segment(line.query_name, line.flag, line.reference_name, reference_start, line.mapping_quality, query_alignment_length, clip, clip_2, pid, line.infer_read_length())
        segment.end = reference_end
        created_subsegments.append(segment)
    return created_subsegments

def read_snp_vcf(contig):
    """Reads vcf/bedfile with pre-set SNP positions and calls find_SNPs() to save the variants"""
    with open(NanoSV.opts_snp_file, "r") as snp_file:
        snp_nummer = 1
        for line in snp_file:
            snp_nummer += 1
            line = line.rstrip()
            if not line.startswith("#"):
                columns = line.split("\t")
                if len(columns) > 1:
            	    if columns[0] == contig:
                        find_SNPs(columns[0], int(columns[1]))

def find_SNPs(chromosome, snp_position):
    """
    Looks for SNPs on given position in the genome using pileup.
    :param chromosome:
    :param snp_position:
    """
    base_ratios = {'A': [0, 0], 'C': [0, 0], 'G': [0, 0], 'T': [0, 0], '=': [0, 0]}
    deletions = 0
    total_n = 0
    variant = v.Variant(chromosome, int(snp_position))
    for pileupcolumn in F.pileup(chromosome, int(snp_position)-1, int(snp_position), truncate=True):
        for pileupread in pileupcolumn.pileups:
            if not keep_segment(pileupread.alignment, pileupread.alignment.query_alignment_length):
                continue
            clip, clip_2 = calculate_clip(pileupread.alignment)
            if pileupread.is_del:
                variant.add_segment([pileupread.alignment.reference_name, pileupread.alignment.reference_start, str(pileupread.alignment.query_name) + ";" + str(clip)], '-')
                deletions += 1
                total_n += 1
            if not pileupread.is_del and not pileupread.is_refskip:
                if pileupread.alignment.query_qualities[pileupread.query_position] >= NanoSV.opts_min_base_qual_ph:
                    base_ratios[pileupread.alignment.query_sequence[pileupread.query_position]][0] += 1
                else:
                    base_ratios[pileupread.alignment.query_sequence[pileupread.query_position]][1] += 1
                variant.add_segment([pileupread.alignment.reference_name, pileupread.alignment.reference_start, str(pileupread.alignment.query_name) + ";" + str(clip)], [pileupread.alignment.query_sequence[pileupread.query_position], pileupread.alignment.query_qualities[pileupread.query_position]])
                total_n += 1
    if deletions < (NanoSV.opts_max_deletions * total_n):
        haplotypes = sorted(base_ratios.items(), key=lambda x: sum(x[1]))[-2:]
        try:
            if haplotypes[0][1][0] / sum(haplotypes[0][1]) > NanoSV.opts_min_occurences_of_highq_var and haplotypes[1][1][0] / sum(haplotypes[1][1]) > NanoSV.opts_min_occurences_of_highq_var:
                if sum(haplotypes[0][1]) / (sum(haplotypes[1][1]) + sum(haplotypes[0][1])) > NanoSV.opts_min_occurences_of_var:
                    bin = int(int(snp_position) / NanoSV.opts_variant_bin_size)
                    variants[chromosome][bin][int(snp_position)] = variant
        except ZeroDivisionError:
                ""

def keep_segment(line, query_alignment_length):
    """
    Calls calculate_pid() and checks if pid meets requirement to use read in NanoSV. If so, returns pid; else False
    :param line:
    :param query_alignment_length:
    :return pid:
    """
    if line.mapping_quality < NanoSV.opts_min_mapq:
        return False
    pid = calculate_pid(line, query_alignment_length)
    if float(pid) < NanoSV.opts_min_pid:
        return False
    return pid

def find_variations_md(cigartuples, chromosome, pos, qual, seq, md, subsegments):
    """
    Loops through read to find variant base. This function is used with BWA-MEM, Minimap2 and NGMLR bam files
    :param cigartuples:
    :param chromosome:
    :param pos:
    :param qual:
    :param seq:
    :param md:
    :param subsegments:
    """
    global tmp_variants
    md_tag = re.split("\^\D+", md)
    read_cursor = 0
    ref_cursor = 0
    md_cursor = 0
    md_string = ''
    for tuple in cigartuples:
        if tuple[0] == 4:
            read_cursor += tuple[1]
        if tuple[0] == 1:
            read_cursor += tuple[1]
        if tuple[0] == 2:
            md_cursor += 1
            md_string = ''
            for deletion in range(tuple[1]):
                if (pos + 1 + ref_cursor) not in tmp_variants:
                    tmp_variants[pos + 1 + ref_cursor] = v.Variant(chromosome, pos + 1 + ref_cursor)
                for segment in subsegments:
                    if segment.pos <= (pos + 1 + ref_cursor) <= segment.end:
                        tmp_variants[pos + 1 + ref_cursor].add_segment(segment.id, ['-'])
                        break
                    if segment.end < ref_cursor:
                        del subsegments[subsegments.index(segment)]
                ref_cursor += 1
        if tuple[0] == 0:
            if re.search("\D+", md_tag[md_cursor]):
                if md_string == '':
                    for m in re.split("\D", md_tag[md_cursor]):
                        if md_string != '':
                            md_string += 'X'
                        md_string += "=" * int(m)
                for seq_mismatch in re.finditer('X', md_string[:tuple[1]]):
                    if (pos + 1 + ref_cursor) not in tmp_variants:
                        tmp_variants[pos + 1 + ref_cursor] = v.Variant(chromosome, pos + 1 + ref_cursor)
                    for segment in subsegments:
                        if segment.pos <= (pos + 1 + ref_cursor) <= segment.end:
                            tmp_variants[pos + 1 + ref_cursor].add_segment(segment.id, [seq[read_cursor + seq_mismatch.start()], qual[read_cursor + seq_mismatch.start()]])
                            break
                        if segment.end < ref_cursor:
                            del subsegments[subsegments.index(segment)]
                md_string = md_string[tuple[1]:]
            ref_cursor += tuple[1]
            read_cursor += tuple[1]

def find_variations_cigar(cigartuples, chromosome, pos, qual, seq, subsegments):
    """
    Loops through read to find variant base. This function is used with LAST bam files
    :param cigartuples:
    :param chromosome:
    :param pos:
    :param qual:
    :param seq:
    :param subsegments:
    """
    global tmp_variants
    ref_cursor = (int(pos))
    read_cursor = 0
    for tuple in cigartuples:
        if tuple[0] == 4:
            read_cursor += tuple[1]
        if tuple[0] == 8:
            for mismatch in range(tuple[1]):
                ref_cursor += 1
                if ref_cursor not in tmp_variants:
                    tmp_variants[ref_cursor] = v.Variant(chromosome, ref_cursor)
                for segment in subsegments:
                    if segment.pos <= ref_cursor <= segment.end:
                        tmp_variants[ref_cursor].add_segment(segment.id, [seq[read_cursor], qual[read_cursor]])
                        break
                    if segment.end < ref_cursor:
                        del subsegments[subsegments.index(segment)]
                read_cursor += 1
        elif tuple[0] == 2:
            for deletion in range(tuple[1]):
                ref_cursor += 1
                if ref_cursor not in tmp_variants:
                    tmp_variants[ref_cursor] = v.Variant(chromosome, ref_cursor)
                for segment in subsegments:
                    if segment.pos <= ref_cursor <= segment.end:
                        tmp_variants[ref_cursor].add_segment(segment.id, ['-'])
                        break
                    if segment.end < ref_cursor:
                        del subsegments[subsegments.index(segment)]
        elif tuple[0] == 7:
            ref_cursor += tuple[1]
            read_cursor += tuple[1]
        elif tuple[0] == 1:
            read_cursor += tuple[1]

def remove_variations(start, end, chr):
    """
    Checks saved variants on position when all reads of that position are parsed. If a variant meets all selection
    requirements it is kept, otherwise deleted
    :param start:
    :param end:
    :param chr:
    """
    global tmp_variants, variants, segments_to_check
    for position in sorted(tmp_variants):
        cov = 0
        for segment in segments_to_check:
            if segments_to_check[segment][0] <= position <= segments_to_check[segment][1]:
                cov += 1
        if cov < NanoSV.opts_minimum_coverage:
            continue
        variant = tmp_variants[position]
        if position < start:
            continue
        elif position > end:
            break
        base_ratios = {'A': [0, 0], 'C': [0, 0], 'G': [0, 0], 'T': [0, 0], '=': [0, 0]}
        deletions = 0
        amount_of_vars = 0
        for segment_id, info in variant.segments.items():
            if info[0] == '-':
                deletions += 1
                amount_of_vars += 1
            elif info[1] >= NanoSV.opts_min_base_qual_ph:
                base_ratios[info[0]][0] += 1
                amount_of_vars += 1
            else:
                base_ratios[info[0]][1] += 1
                amount_of_vars += 1
        base_ratios['='][0] = cov - amount_of_vars
        qual_str = []
        if deletions < (NanoSV.opts_max_deletions * cov):
            haplotypes = sorted(base_ratios.items(), key=lambda x:sum(x[1]))[-2:]
            if haplotypes[0][0] != '=':
                for segment_id, info in variant.segments.items():
                    if info[0] == haplotypes[0][0]:
                        qual_str.append(str(info[1]))
            if haplotypes[1][0] != '=':
                for segment_id, info in variant.segments.items():
                    if info[0] == haplotypes[1][0]:
                        qual_str.append(str(info[1]))
            try:
                if haplotypes[0][1][0]/sum(haplotypes[0][1]) > NanoSV.opts_min_occurences_of_highq_var and haplotypes[1][1][0]/sum(haplotypes[1][1]) > NanoSV.opts_min_occurences_of_highq_var:
                    if sum(haplotypes[0][1])/(sum(haplotypes[1][1]) + sum(haplotypes[0][1])) > NanoSV.opts_min_occurences_of_var:
                        bin = int(position/NanoSV.opts_variant_bin_size)
                        variants[chr][bin][position] = tmp_variants[position]
            except ZeroDivisionError:
                ""
        del tmp_variants[position]

def write_bed():
    """
    Writes all phasing SNP positions to a bed output file.
    """
    with open("snp_positions_used_for_phasing.bed", 'w') as bedfile:
        for chromosome in variants:
            for bin in variants[chromosome]:
                for position, object in variants[chromosome][bin].items():
                    string = str(chromosome) + "\t" + str(position) + "\n"
                    bedfile.write(string)



#!/usr/bin/python
import sys
import pysam
import collections
import math
import os
import vcf as py_vcf

from math import log10
from statistics import median
from version import __version__

from utils import parse_bam as bam
from utils import coverage

sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir))
import NanoSV

class SV:
    def __init__(self, id, breakpoint):
        self.id = id
        self.chr = breakpoint.segment_1["rname"]
        self.chr2 = breakpoint.segment_2["rname"]
        self.flag1 = breakpoint.segment_1["flag"]
        self.flag2 = breakpoint.segment_2["flag"]
        self.pos = [breakpoint.segment_1["pos"]]
        self.ref = 'N'
        self.alt = str(breakpoint.svtype)
        self.qual = None
        self.filter = []
        self.info = {
            'IMPRECISE': True,
            'END': [breakpoint.segment_2["pos"]],
            'SVTYPE': breakpoint.svtype,
            'SVMETHOD': 'NanoSV_v'+__version__,
        }
        self.format = {
            'GT': './.',
            'DV': [1, 1],
            'VO': [(1 - 10 ** (-breakpoint.segment_1["mapq"] / 10.0)),
                   (1 - 10 ** (-breakpoint.segment_2["mapq"] / 10.0))],
            'DR': [0, 0],
            'RO': [0, 0],
            'HR': [0, 0]
        }
        self.breakpoints = [breakpoint.id]
        self.set = 0
        self.SVcluster = 1
        self.ref_qname_clips = [[], []]

    def addBreakpoint(self, breakpoint):
        """
        Adds breakpoint to SV object
        :param breakpoint is breakpoint-object:
        """
        self.breakpoints.append(breakpoint.id)
        self.pos.append(breakpoint.segment_1["pos"])
        self.info['END'].append(breakpoint.segment_2["pos"])
        self.format['DV'][0] += 1
        self.format['DV'][1] += 1
        self.format['VO'][0] += (1 - 10 ** (-breakpoint.segment_1["mapq"] / 10.0))
        self.format['VO'][1] += (1 - 10 ** (-breakpoint.segment_2["mapq"] / 10.0))

    def addInfoField(self, key, value):
        """
        Adds info field to SV object
        :param key is the name of the info field result:
        :param value is the value of the field result:
        """
        if key in self.info:
            if isinstance(self.info[key], list):
                try:
                    if isinstance(self.info[key][0], list) or isinstance(self.info[key][1], list):
                        if value[0] is not None: self.info[key][0].append(value[0][0])
                        if value[1] is not None: self.info[key][1].append(value[1][0])
                    else:
                        self.info[key].append(value[0])

                except IndexError:
                    self.info[key].append(value[0])
        else:
            self.info[key] = value

    def significanceTest(self, avg_dupdel_cov, dup):
        """
        Calculates if there is a significant difference in coverage to call DUP or DEL
        :param avg_dupdel_cov is the distribution with coverage of random positions in the genome:
        :param dup is a boolean. True in case of a duplication, False in case of a deletion:
        :return True or false depending on the outcome of the test. True when there is a significant difference:
        """
        teller = 0
        if dup:
            for value in coverage.coverages:
                if float(value) > avg_dupdel_cov:
                    teller += 1
        else:
            for value in coverage.coverages:
                if float(value) < avg_dupdel_cov:
                    teller += 1
        if teller / len(coverage.coverages) < 0.05:
            self.info["DEPTHPVAL"] = '%.3f' % (teller / len(coverage.coverages))
            return True
        else:
            self.info["DEPTHPVAL"] = '%.3f' % (teller / len(coverage.coverages))
            return False

    def setReferenceBase(self):
        """
        Get the reference bases for the SV object
        :param
        """
        bases = dict()
        bam = pysam.AlignmentFile(NanoSV.opts_bam, 'rb')
        for pileupcolumn in bam.pileup(self.chr,self.pos-1,self.pos, truncate=True):
            for pileupread in pileupcolumn.pileups:
                if not pileupread.is_del and not pileupread.is_refskip:
                    if not pileupread.alignment.query_sequence[pileupread.query_position] in bases:
                        bases[pileupread.alignment.query_sequence[pileupread.query_position]] = 0
                    bases[pileupread.alignment.query_sequence[pileupread.query_position]] += 1
        if (len(bases.keys()) > 0):
            self.ref = sorted(bases, key=bases.__getitem__)[-1]

    def setArguments(self, opts_depth_support):
        """
        Set standard info field arguments.
        :param opts_depth_support is a boolean for if the coverage dup/del analysis is switched on:
        """
        self.info['CIPOS'] = [int(min(self.pos) - median(self.pos)), int(max(self.pos) - median(self.pos))]
        self.info['CIEND'] = [int(min(self.info['END']) - median(self.info['END'])), int(max(self.info['END']) - median(self.info['END']))]
        if self.info['CIPOS'] == [0,0] and self.info['CIEND'] == [0,0]:
            del self.info['IMPRECISE']
        self.pos = int(median(self.pos))
        self.info['END'] = int(median(self.info['END']))
        if self.chr == self.chr2:
            self.info['SVLEN'] = (self.info['END'] - self.pos)
        self.setInfoField()
        self.setReferenceBase()
        dup = 0
        if self.alt == "INS":
            self.alt = py_vcf.model._SV("INS")
        if self.alt == "BND":
            if self.flag1 & 16:
                if self.flag2 & 16:
                    self.alt = py_vcf.model._Breakend(self.chr2, int(self.info['END']), True, False, self.ref, True)
                    if opts_depth_support and self.chr == self.chr2:
                        avg_dupdel_cov = self.getDupDelcoverage()
                        if self.significanceTest(avg_dupdel_cov, True):
                            self.alt = py_vcf.model._SV("DUP")
                            self.info['SVTYPE'] = "DUP"
                        dup = 1
                else:
                    self.alt = py_vcf.model._Breakend(self.chr2, int(self.info['END']), True, True, self.ref, True)
            else:
                if self.flag2 & 16:
                    self.alt = py_vcf.model._Breakend(self.chr2, int(self.info['END']), False, False, self.ref, True)
                else:
                    self.alt = py_vcf.model._Breakend(self.chr2, int(self.info['END']), False, True, self.ref, True)
                    if opts_depth_support and self.chr == self.chr2:
                        avg_dupdel_cov = self.getDupDelcoverage()
                        if self.significanceTest(avg_dupdel_cov, False):
                            self.alt = py_vcf.model._SV("DEL")
                            self.info['SVTYPE'] = "DEL"
        gt_lplist = self.bayes_gt(sum(self.format['RO']), sum(self.format['VO']), dup)
        gt_sum = 0

        pplist = []
        for gt in gt_lplist:
            pplist.append(10 ** gt)
            gt_sum += 10 ** gt

        if gt_sum > 0:
            npplist = [x / gt_sum for x in pplist]

            plplist = []
            for p in npplist:
                if p > 0:
                    plplist.append(int(-10 * log10(p)))
                else:
                    plplist.append(9999)

            gt_idx = plplist.index(min(plplist))
            gq_idx = plplist.index(sorted(plplist)[1])

            plplist[gt_idx] = 0

            if gt_idx == 0:
                self.format['GT'] = '0/0'
            elif gt_idx == 1:
                self.format['GT'] = '0/1'
            elif gt_idx == 2:
                self.format['GT'] = '1/1'

            self.format['PL'] = plplist
            self.format['GQ'] = plplist[gq_idx]
            self.qual = plplist[0]
        else:
            self.format['GT'] = './.'
            self.format['PL'] = [0, 0, 0]
            self.format['GQ'] = [0]
            self.qual = 0

        self.set = 1

    def getDupDelcoverage(self):
        """
        Returns the average coverage of all positions within the start and end of the SV.
        :return average coverage:
        """
        start = round(self.pos)
        stop = round(self.info['END'])
        if start > stop:
            start = round(self.info['END'])
            stop = round(self.pos)
        position = str(self.chr) + ":" + str(start) + "-" + str(stop)
        dupdel_coverages = []

        if 'sambamba' in NanoSV.opts_sambamba:
            cmd = NanoSV.opts_sambamba + " depth base --min-coverage=0 " + NanoSV.opts_bam + " -L " + position + " --nthreads=" + str(NanoSV.opts_threads) + " 2> /dev/null | awk '{if (NR!=1) print $3}'"
        elif 'samtools' in NanoSV.opts_sambamba:
            cmd = NanoSV.opts_sambamba + " depth " + NanoSV.opts_bam + " -r " + position + " | awk '{print $3}'"

        with os.popen(cmd) as commandoutput:
            for line in commandoutput:
                dupdel_coverages.append(float(line.rstrip()))
        return float(sum(dupdel_coverages)) / len(dupdel_coverages)

    def log_choose(self, n, k):
        """
        Returns log of given variable.
        :param n:
        :param k:
        """
        r = 0.0
        # swap for efficiency if k is more than half of n
        if k * 2 > n:
            k = n - k
        for d in range(1, k + 1):
            r += math.log(n, 10)
            r -= math.log(d, 10)
            n -= 1

        return (r)

    def bayes_gt(self, ref, alt, is_dup):
        """
        Returns genotype list
        :param ref:
        :param alt:
        :param is_dup boolean for if SV is dup:
        :return list with probabilities of different genotypes:
        """
        # probability of seeing an alt read with true genotype of of hom_ref, het, hom_alt respectively
        if is_dup:  # specialized logic to handle non-destructive events such as duplications
            p_alt = [1e-3, 0.5, 0.9]
        else:
            p_alt = [1e-3, 0.5, 0.9]

        total = ref + alt
        log_combo = self.log_choose(int(total), int(alt))

        lp_homref = log_combo + alt * math.log(p_alt[0], 10) + ref * math.log(1 - p_alt[0], 10)
        lp_het = log_combo + (alt * math.log(p_alt[1], 10)) + (ref * math.log(1 - p_alt[1], 10))
        lp_homalt = log_combo + alt * math.log(p_alt[2], 10) + ref * math.log(1 - p_alt[2], 10)

        return [lp_homref, lp_het, lp_homalt]

    def setInfoField(self):
        """
        Calculates RT, MAPQ, PID and PLENGTH
        """
        for field in self.info:
            if field == "RT":
                rt = [0, 0, 0]
                for type in self.info[field]:
                    if type == "2d":
                        rt[0] += 1
                    if type == "template":
                        rt[1] += 1
                    if type == "complement":
                        rt[2] += 1

                self.info[field] = rt
            elif isinstance(self.info[field], list):
                if isinstance(self.info[field][0], list):
                    self.info[field][0] = median(map(float, self.info[field][0]))
                    self.info[field][1] = median(map(float, self.info[field][1]))
                    if field == "MAPQ":
                        self.info[field][0] = int(self.info[field][0])
                        self.info[field][1] = int(self.info[field][1])
                    elif field == "PID" or field == "PLENGTH":
                        self.info[field][0] = round(self.info[field][0], 3)
                        self.info[field][1] = round(self.info[field][1], 3)
                else:
                    if 'CI' not in field:
                        self.info[field] = self.median_type(self.info[field])

    def median_type(self, toCalculate):
        """
        Returns median of given list in the original datatype
        :param toCalculate list to calculate median for:
        :return median
        """
        if type(toCalculate[0]) == float:
            return median(toCalculate)
        elif type(toCalculate[0]) == int:
            return int(median(toCalculate))

    def setFormat(self):
        """
        Creates format line of VCF result
        :return a list with the order of the values and the values:
        """
        del self.format['VO']
        del self.format['RO']
        format_list = sorted(self.format.keys())
        format_list.remove('GT')
        format_list.insert(0, 'GT')
        format = ':'.join(format_list)

        data = collections.namedtuple('CallData', format_list)(**self.format)
        call = py_vcf.model._Call('site', bam.sample_name, data)
        samples_indexes = [0]
        samples = [call]
        return [format, samples_indexes, samples]

    def printVCF(self, vcf_writer):
        """
        Combines all the values in a pyvcf record to write to the vcf
        :param vcf_writer is necessary to write a record to the vcf:
        """
        format_output = self.setFormat()
        if isinstance(self.alt, py_vcf.model._Breakend):
            del self.info['END']
        record = py_vcf.model._Record(self.chr, self.pos, self.id, self.ref, [self.alt], self.qual, self.filter, self.info, format_output[0], format_output[1], format_output[2])
        vcf_writer.write_record(record)



import configparser
import sys
import os
import argparse

from version import __version__
import utils

parser = argparse.ArgumentParser()
parser = argparse.ArgumentParser(description='Put here a description.')
parser.add_argument('bam', help='/path/to/file.bam')
parser.add_argument('-t', '--threads', default=4, type=int, help='Number of threads [default: 4]')
parser.add_argument('-s','--sambamba', default='sambamba',type=str,help=' Give the full path to the sambamba or samtools executable [default: sambamba ]')
parser.add_argument('-c','--config', default=os.path.dirname(os.path.abspath(__file__))+"/config.ini",type=str,help='Give the full path to your own ini file [ default: config.ini ]')
parser.add_argument('-b','--bed', default=os.path.dirname(os.path.abspath(__file__))+"/bedfiles/human_hg19.bed",type=str,help=' Give the full path to your own bed file, used for coverage depth calculations [default: human_hg19.bed ]')
parser.add_argument('-o','--output',default=sys.stdout,type=argparse.FileType('w'),help='Give the full path to the output vcf file [default: <stdout> ]')
parser.add_argument('-f', '--snp_file', default=False, type=str, help='Give full path to the SNP variant file for phasing. Supporting file formats: BED and VCF')
parser.add_argument('-v', '--version', action='version',version='NanoSV {}'.format(__version__))

args = parser.parse_args()

opts_bam = args.bam
opts_bed = args.bed
opts_threads = args.threads
opts_sambamba = args.sambamba
opts_output = args.output
opts_snp_file = args.snp_file
cfg = configparser.ConfigParser()
if not args.config == os.path.dirname(os.path.abspath(__file__))+"/config.ini":
    cfg.read([os.path.dirname(os.path.abspath(__file__))+"/config.ini", args.config])
else:
    cfg.read(args.config)

opts_max_split = int(cfg.get('Filter options', 'max_split'))
opts_min_pid = float(cfg.get('Filter options', 'min_pid'))
opts_min_mapq = int(cfg.get('Filter options', 'min_mapq'))

opts_cluster_distance = int(cfg.get('Detection options', 'cluster_distance'))
opts_cluster_count = int(cfg.get('Detection options', 'cluster_count'))
opts_refread_distance = int(cfg.get('Detection options', 'refreads_distance'))
opts_hanging_length = int(cfg.get('Detection options', 'hanging_length'))
opts_mate_distance = int(cfg.get('Detection options', 'mate_distance'))
opts_depth_support = cfg.getboolean('Detection options', 'depth_support')
opts_min_indel_size = str(cfg.get('Detection options', 'min_indel_size'))

opts_qual_flag = int(cfg.get('Output filter options', 'qual_flag'))
opts_window_size = int(cfg.get('Output filter options', 'window_size'))
opts_svcluster = int(cfg.get('Output filter options', 'svcluster'))
opts_mapq_flag = int(cfg.get('Output filter options', 'mapq_flag'))
opts_pid_flag = float(cfg.get('Output filter options', 'pid_flag'))
opts_gap_flag = int(cfg.get('Output filter options', 'gap_flag'))
opts_ci_flag = int(cfg.get('Output filter options', 'ci_flag'))

opts_phasing_on = cfg.getboolean('Phasing Options', 'phasing_on')
opts_variant_bin_size = int(cfg.get('Phasing Options', 'variant_bin_size'))
opts_phasing_window = int(cfg.get('Phasing Options', 'phasing_window'))
opts_min_base_qual_ph = int(cfg.get('Phasing Options', 'min_base_qual_ph'))
opts_clustering_cutoff = float(cfg.get('Phasing Options', 'clustering_cutoff'))
opts_min_occurences_of_var = float(cfg.get('Phasing Options', 'min_occurences_of_var'))
opts_max_deletions = float(cfg.get('Phasing Options', 'max_deletions'))
opts_min_occurences_of_highq_var = float(cfg.get('Phasing Options', 'min_highq_var'))
opts_minimum_coverage = int(cfg.get('Phasing Options', 'min_coverage'))

def main():

    if opts_depth_support:
        utils.coverage.calculate_coverage_bed()
    else:
        coverages = []

    utils.parse_bam.parse_bam()

    utils.create_vcf.print_vcf_header()

    utils.parse_reads.parse_reads()

    utils.parse_breakpoints.parse_breakpoints()

    utils.parse_svs.parse_svs()


if __name__ == "__main__":
    main()



#!/usr/bin/python

import io
import sys
import time
import vcf as py_vcf
import os

from utils import parse_bam as bam
from version import __version__

sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir))
import NanoSV


def print_vcf_header():
    """
    Creates vcf header by setting vcf format and calling functions to create each header section.
    """
    global vcf_writer, vcf

    vcf = py_vcf.Reader(io.StringIO("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t"+ str(bam.sample_name)))
    set_metadata_header(vcf)
    set_contig_header(vcf)
    set_info_header(vcf)
    set_alt_header(vcf)
    set_format_header(vcf)
    set_filter_header(vcf)
    vcf_writer = py_vcf.Writer(NanoSV.opts_output, vcf)


def set_metadata_header(vcf):
    """
    Create metadata header in vcf output file
    :param vcf used to add contigs to file:
    """
    vcf.metadata = {
        'fileformat': 'VCFv4.1',
        'fileDate': str(time.strftime("%c")),
        'cmdline': [" ".join(sys.argv)]
    }


def set_contig_header(vcf):
    """
    Create contig header in vcf output file
    :param vcf used to add contigs to file:
    """
    for item in bam.header["SQ"]:
        vcf.contigs[item['SN']] = py_vcf.parser._Contig(item["SN"], item["LN"])


def set_info_header(vcf):
    """
    Create info header in vcf output file
    :param vcf used to add infos to file:
    """
    vcf.infos = {
        'IMPRECISE': py_vcf.parser._Info("IMPRECISE", 0, "Flag", "Imprecise structural variant", "NanoSV", __version__),
        'SVTYPE': py_vcf.parser._Info("SVTYPE", 1, "String", "Type of structural variant", "NanoSV", __version__),
        'SVMETHOD': py_vcf.parser._Info("SVMETHOD", 1, "String", "Type of approach used to detect SV", "NanoSV", __version__),
        'END': py_vcf.parser._Info("END", 1, "Integer", "End position of structural variant", "NanoSV", __version__),
        'CIPOS': py_vcf.parser._Info("CIPOS", 2, "Integer", "Confidence interval around POS", "NanoSV", __version__),
        'CIEND': py_vcf.parser._Info("CIEND", 2, "Integer", "Confidence interval around END", "NanoSV", __version__),
        'SVLEN': py_vcf.parser._Info("SVLEN", None, "Integer", "Distance between the two genomic positions", "NanoSV", __version__),
        'RT': py_vcf.parser._Info("RT", 3, "Integer", "Number of the different read types (2d, template, complement)","NanoSV", __version__),
        'GAP': py_vcf.parser._Info("GAP", 1, "Integer","Median number of bases between the two segments of the SV, in case of an insertion this is the size of the insertion","NanoSV", __version__),
        'MAPQ': py_vcf.parser._Info("MAPQ", 2, "Integer","Median mapping quality of the two segments of the structural variant", "NanoSV",__version__),
        'PID': py_vcf.parser._Info("PID", 2, "Float","Median percentage identity to the reference of the two segments of the structural variant","NanoSV", __version__),
        'PLENGTH': py_vcf.parser._Info("PLENGTH", 2, "Float","Median segment length percentage of the two segments of the structural variant","NanoSV", __version__),
        'RLENGTH': py_vcf.parser._Info("RLENGTH", 1, "Integer", "Median length of the total reads", "NanoSV", __version__),
        'MATEID': py_vcf.parser._Info('MATEID', None, 'String', 'ID of mate breakend', 'NanoSV', __version__),
        'PURITY_SCORE': py_vcf.parser._Info('PURITY_SCORE', None, "Integer", "Purity of clusters after phasing", "NanoSV", __version__),
        'PHASING_SCORE': py_vcf.parser._Info('PHASING_SCORE', None, "Integer", "Percentage of reads in two largest clusters after phasing", "NanoSV", __version__),
        'SNPS_USED': py_vcf.parser._Info('SNPS_USED', None, "Integer", "SNPs used during phasing", "NanoSV", __version__)
    }

    if NanoSV.opts_depth_support:
        vcf.infos['DEPTHPVAL'] = py_vcf.parser._Info("DEPTHPVAL", 1, "Float", "In case of a duplication or deletion the P-value of the significance test is shown here", "NanoSV", __version__)

def set_format_header(vcf):
    """
    Create format header in vcf output file
    :param vcf used to add formats to file:
    """
    vcf.formats = {
        'GT': py_vcf.parser._Format('GT', 1, 'String', 'Genotype'),
        'DR': py_vcf.parser._Format('DR', 2, 'Integer', 'Number of reference reads'),
        'DV': py_vcf.parser._Format('DV', 2, 'Integer', 'Number of variant reads'),
        'GQ': py_vcf.parser._Format('GQ', 1, 'Integer', 'Genotype quality'),
        'HR': py_vcf.parser._Format('HR', 2, 'Integer', 'Number of hanging variant reads'),
        'PL': py_vcf.parser._Format('PL', 'G', 'Integer','Normalized, Phred-scaled likelihoods for genotypes as defined in the VCF specification')
    }


def set_filter_header(vcf):
    """
    Create filter header in vcf output file
    :param vcf used to add filters to file:
    """
    vcf.filters = {
        'SVcluster': py_vcf.parser._Filter("SVcluster","There are more than " + str(NanoSV.opts_svcluster) + " SVs in a window of " + str(NanoSV.opts_window_size) + " on both sides"),
        'Gap': py_vcf.parser._Filter("Gap", "The median gap size is larger than " + str(NanoSV.opts_gap_flag) + " for non insertions"),
        'MapQual': py_vcf.parser._Filter("MapQual", "The median mapping quality is less than " + str(NanoSV.opts_mapq_flag)),
        'PID': py_vcf.parser._Filter("PID", "The median PID of one of the two sides is less than " + str(NanoSV.opts_pid_flag)),
        'CIPOS': py_vcf.parser._Filter("CIPOS", "The CIPOS is greater than " + str(NanoSV.opts_ci_flag)),
        'CIEND': py_vcf.parser._Filter("CIEND", "The CIEND is greater than " + str(NanoSV.opts_ci_flag)) ,
        'LowQual': py_vcf.parser._Filter('LowQual','QUAL score is less than 20')
    }


def set_alt_header(vcf):
    """
    Create alt header in vcf output file.
    :param vcf used to add alts to file:
    """
    vcf.alts = {
        'DEL': py_vcf.parser._Alt("DEL", "Deletion"),
        'DUP': py_vcf.parser._Alt("DUP", "Duplication"),
        'BND': py_vcf.parser._Alt("BND", "Breakend"),
        'INS': py_vcf.parser._Alt("INS", "Insertion")
    }



#!/usr/bin/python

import os
import time
import sys

sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir))
import NanoSV

coverages = []

def calculate_coverage_bed():
    """
    Create list of coverages based on locations in bed file.
    """
    sys.stderr.write(time.strftime("%c") + " Busy with calculating the coverage distribution...\n")
    if 'sambamba' in NanoSV.opts_sambamba:
        cmd = NanoSV.opts_sambamba + " depth base --min-coverage=0 " + NanoSV.opts_bam + " -L " + NanoSV.opts_bed + " --nthreads=" + str(NanoSV.opts_threads) + " 2> /dev/null | awk '{if (NR!=1) print $3}'"
    elif 'samtools' in NanoSV.opts_sambamba:
        cmd = NanoSV.opts_sambamba + " depth " + NanoSV.opts_bam + " -b " + NanoSV.opts_bed + " | awk '{print $3}'"
    else:
        sys.exit('No sambamba or samtools found')

    with os.popen(cmd) as coverageOutput:
        for coverage in coverageOutput:
            if coverage != "" and coverage != "\n":
                coverages.append(int(coverage))

    if len(coverages) == 0:
        sys.exit("Can't calculate coverage distribution. The bed file may be inappropriate for your bam file.")



#!/usr/bin/python
import sys
import time
import os

from utils import parse_bam as bam
from utils import parse_breakpoints as breakpoint
from utils import phasing as phasing
from utils import create_vcf as c_vcf

sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir))
import NanoSV

def parse_svs():
    """
    Finds reference reads and prepares to print SVs to VCF
    """

    sys.stderr.write(time.strftime("%c") + " Busy with reference reads...\n")
    prev_rname = 0
    for chromosome in sorted(bam.segments):
        for position in sorted(bam.segments[chromosome]):
            for qname_clip in sorted(bam.segments[chromosome][position]):
                if bam.segments[chromosome][position][qname_clip].rname != prev_rname and prev_rname != 0:
                    del breakpoint.structural_variants_region[prev_rname]
                for sv_min in sorted(breakpoint.structural_variants_region[bam.segments[chromosome][position][qname_clip].rname]):
                    if sv_min <= bam.segments[chromosome][position][qname_clip].pos:
                        del breakpoint.structural_variants_region[bam.segments[chromosome][position][qname_clip].rname][sv_min]
                        continue
                    if sv_min > bam.segments[chromosome][position][qname_clip].end:
                        break
                    for sv_max in sorted(breakpoint.structural_variants_region[bam.segments[chromosome][position][qname_clip].rname][sv_min]):
                        if sv_max >= bam.segments[chromosome][position][qname_clip].end:
                            break
                        for sv_id in breakpoint.structural_variants_region[bam.segments[chromosome][position][qname_clip].rname][sv_min][sv_max]:
                            x = breakpoint.structural_variants_region[bam.segments[chromosome][position][qname_clip].rname][sv_min][sv_max][sv_id]
                            if sv_min > (bam.segments[chromosome][position][qname_clip].pos + NanoSV.opts_refread_distance) and \
                                    sv_max < (bam.segments[chromosome][position][qname_clip].end - NanoSV.opts_refread_distance):
                                breakpoint.structural_variants[sv_id].ref_qname_clips[x].append([chromosome, position, qname_clip])
                                breakpoint.structural_variants[sv_id].format['DR'][x] += 1
                                breakpoint.structural_variants[sv_id].format['RO'][x] += (1 - 10 ** (-bam.segments[chromosome][position][qname_clip].mapq / 10.0))
                prev_rname = bam.segments[chromosome][position][qname_clip].rname

    sys.stderr.write(time.strftime("%c") + " Busy with printing to vcf...\n")

    for sv_id in sorted(breakpoint.structural_variants):
        if breakpoint.structural_variants[sv_id].set != 1:
            breakpoint.structural_variants[sv_id].setArguments(NanoSV.opts_depth_support)
        if breakpoint.structural_variants[sv_id].format['GT'] == '0/0':
            continue
        for sv_id_2 in sorted(breakpoint.structural_variants_region_2[breakpoint.structural_variants[sv_id].chr]):
            if sv_id_2 <= sv_id:
                continue
            if breakpoint.structural_variants[sv_id_2].set != 1:
                breakpoint.structural_variants[sv_id_2].setArguments(NanoSV.opts_depth_support)
            if breakpoint.structural_variants[sv_id_2].format['GT'] == '0/0':
                continue
            if breakpoint.structural_variants[sv_id].chr != breakpoint.structural_variants[sv_id_2].chr:
                continue
            if breakpoint.structural_variants[sv_id].chr2 != breakpoint.structural_variants[sv_id_2].chr2:
                continue
            if abs(breakpoint.structural_variants[sv_id].pos - breakpoint.structural_variants[sv_id_2].pos) <= NanoSV.opts_window_size and abs(breakpoint.structural_variants[sv_id].info['END'] - breakpoint.structural_variants[sv_id_2].info['END']) <= NanoSV.opts_window_size:
                breakpoint.structural_variants[sv_id].SVcluster += 1
                breakpoint.structural_variants[sv_id_2].SVcluster += 1
            if breakpoint.structural_variants[sv_id_2].info['SVTYPE'] != "BND":
                continue
            if breakpoint.structural_variants[sv_id].info['SVTYPE'] != "BND":
                continue
            if (not breakpoint.structural_variants[sv_id].flag1 & 16 and not breakpoint.structural_variants[sv_id].flag2 & 16) and not (breakpoint.structural_variants[sv_id_2].flag1 & 16 and breakpoint.structural_variants[sv_id_2].flag2 & 16):
                continue
            if (not breakpoint.structural_variants[sv_id].flag1 & 16 and breakpoint.structural_variants[sv_id].flag2 & 16) and not (breakpoint.structural_variants[sv_id_2].flag1 & 16 and not breakpoint.structural_variants[sv_id_2].flag2 & 16):
                continue
            if (breakpoint.structural_variants[sv_id].flag1 & 16 and not breakpoint.structural_variants[sv_id].flag2 & 16) and not (not breakpoint.structural_variants[sv_id_2].flag1 & 16 and breakpoint.structural_variants[sv_id_2].flag2 & 16):
                continue
            if (breakpoint.structural_variants[sv_id].flag1 & 16 and breakpoint.structural_variants[sv_id].flag2 & 16) and not (not breakpoint.structural_variants[sv_id_2].flag1 & 16 and not breakpoint.structural_variants[sv_id_2].flag2 & 16):
                continue
            if abs(breakpoint.structural_variants[sv_id].pos - breakpoint.structural_variants[sv_id_2].pos) > NanoSV.opts_mate_distance:
                continue
            if abs(breakpoint.structural_variants[sv_id].info['END'] - breakpoint.structural_variants[sv_id_2].info['END']) > NanoSV.opts_mate_distance:
                continue
            breakpoint.structural_variants[sv_id].info['MATEID'] = sv_id_2
            breakpoint.structural_variants[sv_id_2].info['MATEID'] = sv_id
        if breakpoint.structural_variants[sv_id].qual < NanoSV.opts_qual_flag:
            breakpoint.structural_variants[sv_id].filter.append("LowQual")
        if breakpoint.structural_variants[sv_id].SVcluster > NanoSV.opts_svcluster:
            breakpoint.structural_variants[sv_id].filter.append("SVcluster")
        if breakpoint.structural_variants[sv_id].info['GAP'] > NanoSV.opts_gap_flag and breakpoint.structural_variants[sv_id].info['SVTYPE'] != "INS":
            breakpoint.structural_variants[sv_id].filter.append("GAP")
        if breakpoint.structural_variants[sv_id].info['MAPQ'][0] < NanoSV.opts_mapq_flag or breakpoint.structural_variants[sv_id].info['MAPQ'][1] < NanoSV.opts_mapq_flag:
            breakpoint.structural_variants[sv_id].filter.append("MapQual")
        if breakpoint.structural_variants[sv_id].info['PID'][0] < NanoSV.opts_pid_flag or breakpoint.structural_variants[sv_id].info['PID'][1] < NanoSV.opts_pid_flag:
            breakpoint.structural_variants[sv_id].filter.append("PID")
        if abs(breakpoint.structural_variants[sv_id].info['CIPOS'][0])+breakpoint.structural_variants[sv_id].info['CIPOS'][1] > NanoSV.opts_ci_flag:
            breakpoint.structural_variants[sv_id].filter.append("CIPOS")
        if abs(breakpoint.structural_variants[sv_id].info['CIEND'][0])+breakpoint.structural_variants[sv_id].info['CIEND'][1] > NanoSV.opts_ci_flag:
            breakpoint.structural_variants[sv_id].filter.append("CIEND")
        if breakpoint.structural_variants[sv_id].info['SVTYPE'] == 'INS':
            breakpoint.structural_variants[sv_id].info['SVLEN'] = breakpoint.structural_variants[sv_id].info['GAP']

        windows = []
        if breakpoint.structural_variants[sv_id].flag1 == 16:
            windows.append([breakpoint.structural_variants[sv_id].pos, breakpoint.structural_variants[sv_id].pos + NanoSV.opts_phasing_window])
        elif breakpoint.structural_variants[sv_id].flag1 == 0:
            windows.append([breakpoint.structural_variants[sv_id].pos - NanoSV.opts_phasing_window, breakpoint.structural_variants[sv_id].pos])
        if breakpoint.structural_variants[sv_id].flag2 == 0:
            windows.append([breakpoint.structural_variants[sv_id].info['END'], breakpoint.structural_variants[sv_id].info['END'] + NanoSV.opts_phasing_window])
        elif breakpoint.structural_variants[sv_id].flag2 == 16:
            windows.append([breakpoint.structural_variants[sv_id].info['END'] - NanoSV.opts_phasing_window, breakpoint.structural_variants[sv_id].info['END']])
        if breakpoint.structural_variants[sv_id].format['GT'] == '0/1' and NanoSV.opts_phasing_on:
            phasing_result = phasing.make_matrix(sv_id, windows)
            breakpoint.structural_variants[sv_id].info['PURITY_SCORE'] = phasing_result[0]
            breakpoint.structural_variants[sv_id].info['PHASING_SCORE'] = phasing_result[1]
            breakpoint.structural_variants[sv_id].info['SNPS_USED'] = phasing_result[3]
            breakpoint.structural_variants[sv_id].info['PHASING_PVALUE'] = phasing_result[2]
        breakpoint.structural_variants[sv_id].printVCF(c_vcf.vcf_writer)



#!/usr/bin/python

class Read:
    def __init__(self, qname, length):
        self.qname = qname
        self.length = length
        self.segments = dict()

    def addSegment(self, segment):
        """
        Adds segment to read
        :param segment:
        """
        self.segments[segment.clip] = [segment.rname, segment.pos]



from utils import parse_breakpoints as breakpoint
from utils import parse_reads as read
from utils import parse_bam as bam
from utils import create_vcf as c_vcf
import NanoSV
import random
matrix = []

def make_matrix(sv_id, windows):
    """
    Create matrix of positions and reads with their ref/alt/- as variable. Parse phasing result and select
    best result
    :param sv_id:
    :param windows:
    :return best result with the purity, phasing score and snps used:
    """
    global matrix
    x = 0
    scores = [0,0,0,0]
    bp = -1
    for window in windows:
        matrix = []
        if x == 0:
            chr = breakpoint.structural_variants[sv_id].chr
        else:
            chr = breakpoint.structural_variants[sv_id].chr2
        bin_start = int(window[0] / NanoSV.opts_variant_bin_size)
        bin_end = int(window[1] / NanoSV.opts_variant_bin_size)
        if bin_start < 0:
            bin_start = 0
        if bin_end > c_vcf.vcf.contigs[chr][1]:
            bin_end = int(c_vcf.vcf.contigs[chr][1] / NanoSV.opts_variant_bin_size)
        sv_reads = get_sv_reads(sv_id, x)
        for bin in range(bin_start, bin_end+1):
            for variant_position in bam.variants[chr][bin]:
                if int(window[0]) <= variant_position <= int(window[1]):
                    matrix.append([])
                    matrix_fill_position(breakpoint.structural_variants[sv_id].ref_qname_clips[x], variant_position, sv_id, chr, bin, x)
                    matrix_fill_position(sv_reads, variant_position, sv_id, chr, bin, x)
        matrix = list(map(list, zip(*matrix)))
        if len(matrix) == 0:
            x += 1
            continue
        phasing_result= clustering(matrix, list(range(len(breakpoint.structural_variants[sv_id].ref_qname_clips[x]),len(matrix))), x)
        if sum(phasing_result[:1]) > sum(scores[:1]):
            scores = phasing_result
            scores.append(len(matrix[0]))
            bp += 1
        x += 1
        deleted = 0
        for segment in range(len(matrix)):
            if len(set(matrix[segment-deleted])) <= 1 and matrix[segment-deleted][0] == "-":
                del matrix[segment-deleted]
                deleted += 1
    return scores


def get_sv_reads(sv_id, x):
    """
    gets reads supporting the sv. Used later to evaluate clustering
    :param sv_id:
    :param x:
    :return list with sv_reads:
    """
    sv_reads = []
    for bp_id in breakpoint.structural_variants[sv_id].breakpoints:
        if x == 0:
            sv_reads.append(read.breakpoints[bp_id].segment_1['id'])
        else:
            sv_reads.append(read.breakpoints[bp_id].segment_2['id'])
    return sv_reads


def matrix_fill_position(sv_or_ref, variant_position, sv_id, chr, bin, x):
    """
    fill position in matrix with correct value for either ref or alt base, or a - with unknown haplotype
    :param sv_or_ref:
    :param variant_position:
    :param sv_id:
    :param chr:
    :param bin:
    :param x:
    """
    global matrix
    for segment_id in sv_or_ref:
        if segment_id[2] in bam.variants[chr][bin][variant_position].segments:
            matrix[len(matrix) - 1].append(bam.variants[chr][bin][variant_position].segments[segment_id[2]][0])
        else:
            if breakpoint.structural_variants[sv_id].flag1 == 0 and x == 0:
                if bam.segments[segment_id[0]][segment_id[1]][segment_id[2]].pos > variant_position:
                    matrix[len(matrix) - 1].append('-')
                else:
                    matrix[len(matrix) - 1].append('=')
            elif breakpoint.structural_variants[sv_id].flag1 == 16 and x == 0:
                if bam.segments[segment_id[0]][segment_id[1]][segment_id[2]].end < variant_position:
                    matrix[len(matrix) - 1].append('-')
                else:
                    matrix[len(matrix) - 1].append('=')
            elif breakpoint.structural_variants[sv_id].flag2 == 0 and x == 1:
                if bam.segments[segment_id[0]][segment_id[1]][segment_id[2]].end < variant_position:
                    matrix[len(matrix) - 1].append('-')
                else:
                    matrix[len(matrix) - 1].append('=')
            elif breakpoint.structural_variants[sv_id].flag2 == 16 and x == 1:
                if bam.segments[segment_id[0]][segment_id[1]][segment_id[2]].pos > variant_position:
                    matrix[len(matrix) - 1].append('-')
                else:
                    matrix[len(matrix) - 1].append('=')


def clustering(matrix, sv_reads, bp_id):
    """
    creates clustering matrix by calling make_clustering_matrix(). Clusters most similar reads and sends
    result to judge_clustering to be evaluated. Result comes back and is returned.
    :param matrix:
    :param sv_reads:
    :param bp_id:
    :return phasing result:
    """
    clustering_matrix = make_clustering_matrix(matrix)
    while len(clustering_matrix) > 2:
        keys = []
        for x in clustering_matrix:
            keys.append(x)
        highest_score = 0
        readA = 0
        readB = 0
        for i in keys:
            for key, value in clustering_matrix[i].items():
                if value >= highest_score:
                    highest_score = value
                    readA = key
                    readB = i
        if highest_score < NanoSV.opts_clustering_cutoff:
            break
        merged_name = str(readA) + "," + str(readB)
        merged_dict = {}
        for j in keys:
            if j == readA or j == readB:
                continue
            sum_of_scores = 0
            for read in [readA, readB]:
                if max(map(int, read.split(","))) >= max(map(int, j.split(","))):
                    sum_of_scores += clustering_matrix[str(read)][str(j)]
                else:
                    sum_of_scores += clustering_matrix[str(j)][str(read)]
            merged_dict[str(j)] = sum_of_scores / 2
        del_list = []
        for item, merged_value in merged_dict.items():
            if max(map(int, item.split(","))) <= max(map(int, readB.split(","))):
                continue
            clustering_matrix[str(item)][str(merged_name)] = merged_value
            del_list.append(item)
        del clustering_matrix[str(readA)]
        del clustering_matrix[str(readB)]
        for read in clustering_matrix:
            if readA in clustering_matrix[read]:
                del clustering_matrix[read][readA]
            if readB in clustering_matrix[read]:
                del clustering_matrix[read][readB]
        for item in del_list:
            del merged_dict[str(item)]
        clustering_matrix[merged_name] = merged_dict
    breakpoint_result = judge_clustering(clustering_matrix, sv_reads, len(matrix), bp_id)
    return breakpoint_result


def judge_clustering(clustering_matrix, sv_reads, total_reads, x):
    """
    Judges clustering result and calculates purity and phasing scores. Also calls randomise() 100000 times to get a
    random result. With these random results a p-value can be calculated.
    :param clustering_matrix:
    :param sv_reads:
    :param total_reads:
    :param x:
    :return list with purity score, phasing score and p-value:
    """
    purity_chart = []
    phasing_chart = []
    clusters = []
    for key in clustering_matrix:
        clusters.append(key.split(","))
    longest_clusters = []
    for length_cluster in [len(clusters), len(clusters)-1]:
        long_cluster = []
        for j in range(length_cluster):
            if len(long_cluster) <= len(clusters[j]):
                long_cluster = clusters[j]
        del clusters[clusters.index(long_cluster)]
        longest_clusters.append(long_cluster)
    amounts_per_cluster = []
    clusternr = 0
    for cluster in longest_clusters:
        amounts = [0, 0]
        clusternr += 1
        for read in cluster:
            if int(read) in sv_reads:
                amounts[0] += 1
            else:
                amounts[1] += 1
        amounts_per_cluster.append(amounts)
    pur_percentage_per_cluster = []
    phasing_percentage_per_cluster = []
    for i in range(len(amounts_per_cluster)):
        purity_percentages = []
        purity_percentages.append(amounts_per_cluster[i][0] / (amounts_per_cluster[i][0] + amounts_per_cluster[i][1]) * 100)
        purity_percentages.append(amounts_per_cluster[i][1] / (amounts_per_cluster[i][0] + amounts_per_cluster[i][1]) * 100)
        pur_percentage_per_cluster.append(purity_percentages)
        phasing_percentages = []
        phasing_percentages.append(amounts_per_cluster[i][0] / len(sv_reads) * 100)
        try:
            phasing_percentages.append(amounts_per_cluster[i][1] / (total_reads - len(sv_reads)) * 100)
        except ZeroDivisionError:
            phasing_percentages.append(0)
        phasing_percentage_per_cluster.append(phasing_percentages)


    pur_sv_score = (pur_percentage_per_cluster[0][0] - pur_percentage_per_cluster[1][0])
    pur_ref_score = (pur_percentage_per_cluster[0][1] - pur_percentage_per_cluster[1][1])
    if pur_sv_score < 0:
        pur_sv_score = pur_sv_score * -1
    if pur_ref_score < 0:
        pur_ref_score = pur_ref_score * -1
    phasing_sv_score = (phasing_percentage_per_cluster[0][0] + phasing_percentage_per_cluster[1][0])
    phasing_ref_score = (phasing_percentage_per_cluster[0][1] + phasing_percentage_per_cluster[1][1])

    purity_score = (pur_sv_score + pur_ref_score) / 2
    phasing_score = (phasing_sv_score + phasing_ref_score) / 2
    random_purities = []
    for i in range(100000):
        random_purities.append(randomise(longest_clusters, sv_reads))
    teller = 1
    for value in random_purities:
        if float(value) >= purity_score:
            teller += 1
    pvalue = teller/len(random_purities)
    purity_chart.append(int(purity_score))
    phasing_chart.append(int(phasing_score))
    return [int(purity_score), int(phasing_score), pvalue]


def randomise(longest_clusters, sv_reads):
    """
    Creates a random result and calculates scores to be used with a calculation of the p-value.
    :param longest_clusters:
    :param sv_reads:
    :return purity score:
    """
    random_options = []
    for ref in range(len(matrix) - len(sv_reads)):
        random_options.append(1)
    for alt in range(len(sv_reads)):
        random_options.append(2)
    random.shuffle(random_options)
    new_clusters = [[], [], []]
    for clusternr in range(0,3):
        if clusternr == 0:
            for zero in range(len(matrix) - (len(longest_clusters[0])+len(longest_clusters[1]))):
                new_clusters[clusternr].append(random_options[-1])
                del random_options[-1]
        else:
            for one_or_two in range(len(longest_clusters[clusternr-1])):
                new_clusters[clusternr].append(random_options[-1])
                del random_options[-1]

    pur_percentage_per_cluster = []
    amounts_per_cluster = [[new_clusters[1].count(2),new_clusters[1].count(1)], [new_clusters[2].count(2), new_clusters[2].count(1)]]
    for i in range(len(amounts_per_cluster)):
        purity_percentages = []
        purity_percentages.append(amounts_per_cluster[i][0] / (amounts_per_cluster[i][0] + amounts_per_cluster[i][1]) * 100)
        purity_percentages.append(amounts_per_cluster[i][1] / (amounts_per_cluster[i][0] + amounts_per_cluster[i][1]) * 100)
        pur_percentage_per_cluster.append(purity_percentages)
    pur_sv_score = (pur_percentage_per_cluster[0][0] - pur_percentage_per_cluster[1][0])
    pur_ref_score = (pur_percentage_per_cluster[0][1] - pur_percentage_per_cluster[1][1])
    if pur_sv_score < 0:
        pur_sv_score = pur_sv_score * -1
    if pur_ref_score < 0:
        pur_ref_score = pur_ref_score * -1

    purity_score = (pur_sv_score + pur_ref_score) / 2
    return purity_score


def make_clustering_matrix(matrix):
    """
    calculates similarity between reads to be put in the clustering matrix
    :param matrix:
    :return filled clustering matrix:
    """
    clustering_matrix = {}
    for i in range(len(matrix)):
        clustering_matrix[str(i)] = {}
        for j in range(i + 1):
            mutations_in_common = 0
            if j == i:
                clustering_matrix[str(i)][str(j)] = 0
            else:
                amount_positions = len(matrix[i])
                for pos in range(len(matrix[i])):
                    if matrix[i][pos] == '-' and matrix[j][pos] == '-':
                        amount_positions -= 1
                        continue
                    if matrix[i][pos] == matrix[j][pos]:
                        mutations_in_common += 1
                if amount_positions == 0:
                    clustering_matrix[str(i)][str(j)] = 0
                else:
                    clustering_matrix[str(i)][str(j)] = mutations_in_common / amount_positions
    return clustering_matrix



#!/usr/bin/python


class Segment:
    def __init__(self, qname, flag, rname, pos, mapq, length, clip, clip_2, pid, rlength):
        self.id = [rname, pos, qname + ";" + str(clip)]
        self.qname = qname
        self.rlength = rlength
        self.flag = int(flag)
        self.rname = rname
        self.pos = int(pos)
        self.mapq = int(mapq)
        self.length = int(length)
        self.end = (int(pos) + int(length))
        self.clip = clip
        self.clip_2 = clip_2
        self.pid = pid

    def setPlength(self, rlength):
        """
        Calculates plength
        :param rlength is used to calculate the plength:
        """
        self.plength = format(self.length / rlength, '.3f')



#!/usr/bin/python
import re
import sys
import time
import os

from collections import defaultdict

from classes import sv as svclass
from utils import parse_reads as read
from utils import parse_bam as bam

sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir))
import NanoSV

svID = 1
structural_variants = {}
structural_variants_region = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(int))))
structural_variants_region_2 = defaultdict(lambda: defaultdict(int))


def addSVInfo(sv):
    """
    Adds info to the newly made SV object
    :param sv is SV-object:
    """
    if sum(sv.format['DV']) >= NanoSV.opts_cluster_count * 2:
        flag_list = [sv.flag1, sv.flag2]
        for flag_idx in range(len(flag_list)):
            if (flag_idx == 0 and not flag_list[flag_idx] & 16) or (flag_idx == 1 and flag_list[flag_idx] & 16):
                head_tail = 'T'
            else:
                head_tail = 'H'

            for hanging_pos in sorted(read.hanging_breakpoints_region[sv.chr][head_tail]):
                if hanging_pos < (min(sv.pos) - NanoSV.opts_cluster_distance):
                    continue
                if hanging_pos > (max(sv.pos) + NanoSV.opts_cluster_distance):
                    break
                for hanging_id in read.hanging_breakpoints_region[sv.chr][head_tail][hanging_pos]:
                    segment_id = read.hanging_breakpoints_region[sv.chr][head_tail][hanging_pos][hanging_id]
                    sv.format['HR'][flag_idx] += 1
                    sv.format['VO'][flag_idx] += (1 - 10 ** (-bam.segments[segment_id[0]][segment_id[1]][segment_id[2]].mapq / 10.0))
                    sv.addInfoField("PID", [[bam.segments[segment_id[0]][segment_id[1]][segment_id[2]].pid], None])
                    sv.addInfoField("MAPQ", [[bam.segments[segment_id[0]][segment_id[1]][segment_id[2]].mapq], None])
                    sv.addInfoField("PLENGTH", [[bam.segments[segment_id[0]][segment_id[1]][segment_id[2]].plength], None])
                    sv.addInfoField("RLENGTH", [bam.reads[bam.segments[segment_id[0]][segment_id[1]][segment_id[2]].qname].length])
                    if re.match("/2D_2d$/", bam.segments[segment_id[0]][segment_id[1]][segment_id[2]].qname):
                        sv.addInfoField("RT", ["2d"])
                    elif re.match("/2D_complement$/", bam.segments[segment_id[0]][segment_id[1]][segment_id[2]].qname):
                        sv.addInfoField("RT", ["complement"])
                    else:
                        sv.addInfoField("RT", ["template"])
        structural_variants[sv.id] = sv
        structural_variants_region[sv.chr][min(sv.pos)][max(sv.pos)][sv.id] = 0
        structural_variants_region[sv.chr2][min(sv.info['END'])][max(sv.info['END'])][sv.id] = 1
        structural_variants_region_2[sv.chr][sv.id] = 1


def parse_breakpoints_2(breakpoints_region_2):
    """
    Creates an SV object for every SV.
    :param breakpoints_region_2:
    """
    global svID
    prev_pos_2 = -1
    for pos_2 in sorted(breakpoints_region_2):
        for pos_1 in sorted(breakpoints_region_2[pos_2]):
            for breakpoint_id in breakpoints_region_2[pos_2][pos_1]:
                breakpoint = read.breakpoints[breakpoint_id]
                if (prev_pos_2 == -1):
                    sv = svclass.SV(svID, breakpoint)
                    svID += 1
                elif abs(breakpoint.segment_2["pos"] - prev_pos_2) <= NanoSV.opts_cluster_distance:
                    sv.addBreakpoint(breakpoint)
                else:
                    addSVInfo(sv)
                    sv = svclass.SV(svID, breakpoint)
                    svID += 1
                segment_id1 = breakpoint.segment_1["id"]
                segment_id2 = breakpoint.segment_2["id"]
                sv.addInfoField("PID",[ [bam.segments[segment_id1[0]][segment_id1[1]][segment_id1[2]].pid],[bam.segments[segment_id2[0]][segment_id2[1]][segment_id2[2]].pid] ])
                sv.addInfoField("MAPQ",[ [bam.segments[segment_id1[0]][segment_id1[1]][segment_id1[2]].mapq],[bam.segments[segment_id2[0]][segment_id2[1]][segment_id2[2]].mapq] ])
                sv.addInfoField("PLENGTH",[ [bam.segments[segment_id1[0]][segment_id1[1]][segment_id1[2]].plength],[bam.segments[segment_id2[0]][segment_id2[1]][segment_id2[2]].plength] ])
                sv.addInfoField("RLENGTH",[ bam.reads[bam.segments[segment_id2[0]][segment_id2[1]][segment_id2[2]].qname].length ])
                sv.addInfoField("GAP", [breakpoint.gap])

                if re.match("/2D_2d$/", bam.segments[segment_id2[0]][segment_id2[1]][segment_id2[2]].qname):
                    sv.addInfoField("RT", ["2d"])
                elif re.match("/2D_complement$/", bam.segments[segment_id2[0]][segment_id2[1]][segment_id2[2]].qname):
                    sv.addInfoField("RT", ["complement"])
                else:
                    sv.addInfoField("RT", ["template"])
                prev_pos_2 = int(pos_2)
    addSVInfo(sv)


def parse_breakpoints():
    """
    Loops through breakpoints_region dict and calls parse_breakpoints_2
    """
    sys.stderr.write(time.strftime("%c") + " Busy with parsing breakpoints...\n")

    for region in read.breakpoints_region:
        prev_pos_1 = -1
        breakpoints_region_2 = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))
        for pos_1 in sorted(read.breakpoints_region[region]):
            for breakpoint_id in read.breakpoints_region[region][pos_1]:
                if prev_pos_1 == -1:
                    breakpoints_region_2[read.breakpoints[breakpoint_id].segment_2["pos"]][pos_1][breakpoint_id] = 1
                elif abs(pos_1 - prev_pos_1) <= NanoSV.opts_cluster_distance:
                    breakpoints_region_2[read.breakpoints[breakpoint_id].segment_2["pos"]][pos_1][breakpoint_id] = 1
                else:
                    parse_breakpoints_2(breakpoints_region_2)
                    breakpoints_region_2.clear()
                    breakpoints_region_2[read.breakpoints[breakpoint_id].segment_2["pos"]][pos_1][breakpoint_id] = 1
                prev_pos_1 = int(pos_1)
        parse_breakpoints_2(breakpoints_region_2)

    return structural_variants, structural_variants_region, structural_variants_region_2



#!/usr/bin/python
import sys
import time
import os

from collections import defaultdict

from classes import breakpoint as b
from utils import parse_bam as bam

sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir))
import NanoSV

breakpoints = {}
breakpoints_region = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))
hanging_breakpoints_region = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(int))))

def parse_reads():
    """
    Loops through all reads and saves breakpoints in objects of the Breakpoint class
    """
    sys.stderr.write(time.strftime("%c") + " Busy with parsing read segments...\n")
    breakpointID = 1
    hanging_breakpointID = -1
    for qname in bam.reads:
        clips = sorted(bam.reads[qname].segments)
        if len(clips) == 1 or len(clips) > NanoSV.opts_max_split:
            continue
        for i in range(0, (len(clips) - 1)):
            i2 = i + 1
            rname1, pos1 = bam.reads[qname].segments[clips[i]]
            segment_1 = bam.segments[rname1][pos1][qname + ";" + str(clips[i])]
            rname2, pos2 = bam.reads[qname].segments[clips[i2]]
            segment_2 = bam.segments[rname2][pos2][qname + ";" + str(clips[i2])]
            segment_1.setPlength(bam.reads[qname].length)
            segment_2.setPlength(bam.reads[qname].length)
            breakpoint = b.Breakpoint(breakpointID, segment_1, segment_2)
            breakpointID += 1
            gap = (clips[i2] - (clips[i] + segment_1.length))
            breakpoint.setGap(gap)
            breakpoint.setBreakpoint(segment_1, segment_2)
            if segment_1.rname > segment_2.rname:
                breakpoint.switchSegments()
            elif segment_1.rname == segment_2.rname and breakpoint.segment_1["pos"] > breakpoint.segment_2["pos"]:
                breakpoint.switchSegments()
            breakpoint.setSVtype()
            breakpoints[breakpoint.id] = breakpoint
            values = (breakpoint.svtype, str(breakpoint.segment_1["rname"]), str(breakpoint.segment_2["rname"]), str(breakpoint.segment_1["flag"]), str(breakpoint.segment_2["flag"]))
            breakpoints_region["\t".join(values)][breakpoint.segment_1["pos"]][breakpoint.id] = 1
            if i == 0 and segment_1.clip >= NanoSV.opts_hanging_length:
                hanging_breakpoint_pos = segment_1.pos
                if segment_1.flag & 16:
                    hanging_breakpoint_pos = segment_1.end
                    hanging_breakpoints_region[segment_1.rname]['T'][hanging_breakpoint_pos][hanging_breakpointID] = segment_1.id
                    hanging_breakpointID -= 1
                else:
                    hanging_breakpoints_region[segment_1.rname]['H'][hanging_breakpoint_pos][hanging_breakpointID] = segment_1.id
                    hanging_breakpointID = hanging_breakpointID - 1
            if i2 == len(clips) - 1 and segment_2.clip_2 >= NanoSV.opts_hanging_length:
                hanging_breakpoint_pos = segment_2.end
                if segment_2.flag & 16:
                    hanging_breakpoint_pos = segment_2.pos
                    hanging_breakpoints_region[segment_2.rname]['H'][hanging_breakpoint_pos][hanging_breakpointID] = segment_2.id
                    hanging_breakpointID -= 1
                else:
                    hanging_breakpoints_region[segment_2.rname]['T'][hanging_breakpoint_pos][hanging_breakpointID] = segment_2.id
                    hanging_breakpointID -= 1



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))
exec(open('nanosv/version.py').read())

def readme():
    with open('README.rst') as f:
        return f.read()


setup( name='NanoSV',
       version=__version__,
       description='Structural variation detection tool for Oxford Nanopore data.',
       long_description=readme(),
       classifiers=[
           'Development Status :: 4 - Beta',
           'Intended Audience :: Science/Research',
           'License :: OSI Approved :: MIT License',
           'Programming Language :: Python :: 3',
           'Topic :: Scientific/Engineering :: Bio-Informatics'
       ],
       keywords='nanosv sv-caller nanopore',
       url='https://github.com/mroosmalen/nanosv',
       author='Mark van Roosmalen',
       author_email='m.vanroosmalen-2@umcutrecht.nl',
       license='MIT',
       python_requires='>=3',
       install_requires=['pysam>=0.11','pyvcf','configparser'],
       packages=['nanosv','nanosv.utils','nanosv.classes'],
       package_data={'nanosv': ['config.ini','bedfiles/*.bed']},
       including_package_data=True,
       entry_points={
           'console_scripts': [
               'NanoSV=nanosv.NanoSV:main'
           ]
       }
)



#!/usr/bin/python

class Breakpoint:
    def __init__(self, id, segment_1, segment_2):
        self.id = id
        if segment_1.flag & 16:
            flag_1 = 16
        else:
            flag_1 = 0
        if segment_2.flag & 16:
            flag_2 = 16
        else:
            flag_2 = 0
        self.segment_1 = {
            "id": segment_1.id,
            "rname": segment_1.rname,
            "flag": flag_1,
            "mapq": segment_1.mapq
        }
        self.segment_2 = {
            "id": segment_2.id,
            "rname": segment_2.rname,
            "flag": flag_2,
            "mapq": segment_2.mapq
        }

    def setGap(self, gap):
        """
        Saves gap in class
        :param gap:
        """
        self.gap = gap

    def setBreakpoint(self, segment_1, segment_2):
        """
        Saves start positions of segments depending on the flag
        :param segment_1 used to access and save data:
        :param segment_2 used to access and save data:
        """
        if segment_1.flag & 16:
            self.segment_1["pos"] = segment_1.pos
        else:
            self.segment_1["pos"] = segment_1.end

        if segment_2.flag & 16:
            self.segment_2["pos"] = segment_2.end
        else:
            self.segment_2["pos"] = segment_2.pos

    def switchSegments(self):
        """
        Switches segments in case segment 2 comes before segment 1 in the genome
        """
        self.segment_1, self.segment_2 = self.segment_2, self.segment_1
        self.segment_1["flag"] = 0 if self.segment_1["flag"] & 16 else 16
        self.segment_2["flag"] = 0 if self.segment_2["flag"] & 16 else 16

    def setSVtype(self):
        """
        Sets SV-type to INS in case the difference between the two segments is smaller than the gap setting
        """
        self.svtype = "BND"
        if abs(self.segment_2["pos"] - self.segment_1["pos"]) < self.gap:
            self.svtype = "INS"
            self.segment_2["pos"] = self.segment_1["pos"]+1
            self.segment_1["flag"] = 0
            self.segment_2["flag"] = 0




class Variant:
    def __init__(self, chr, pos):
        self.chr = chr
        self.pos = pos
        self.segments = {}

    def add_segment(self, segment_id, variant_info):
        """
        Adds segment to variant object that contains a similar variant
        :param segment_id: 
        :param variant_info: 
        :return: 
        """
        self.segments[segment_id[2]] = variant_info
